# (c) Copyright 2018 Trent Hauck
# All Rights Reserved
"""A Tokenizer that works on biological sequences."""

from typing import List, Optional, Dict, cast
from dataclasses import field
import itertools as it

from pydantic import dataclasses


@dataclasses.dataclass
class Vocab:
    """A vocabulary object."""

    token_to_int: Dict[str, int]
    int_to_token: Dict[int, str]

    def __len__(self) -> int:
        """Return the length of the vocab."""
        return len(self.token_to_int)

    @classmethod
    def from_list(cls, tokens: List[str]) -> "Vocab":
        """Create a vocabulary from a list of tokens."""
        token_to_int = {}
        int_to_token = {}

        for i, token in enumerate(tokens):
            int_to_token[i] = token
            token_to_int[token] = i

        return cls(token_to_int, int_to_token)


@dataclasses.dataclass
class SequenceTokenizerSpec:
    """The specification for the tokenizer."""

    max_length: int
    alphabet: str

    vocabulary: Vocab = field(init=False)

    kmer_size: int = 1
    kmer_step_size: int = 1

    bos_token: Optional[str] = None
    eos_token: Optional[str] = None
    unk_token: Optional[str] = None
    pad_token: Optional[str] = None
    mask_token: Optional[str] = None

    def __post_init__(self):
        """Post inits the tokenizer spec."""
        self.vocabulary = Vocab.from_list(self.special_tokens + self.possible_kmers)

    @property
    def possible_kmers(self) -> List[str]:
        """Return the set of possible kmers given the alphabet and kmer size."""
        return ["".join(kmer) for kmer in it.product(self.alphabet, repeat=self.kmer_size)]

    @property
    def special_tokens(self) -> List[str]:
        """Return the list of special tokens."""
        return [
            s
            for s in [
                self.bos_token,
                self.eos_token,
                self.unk_token,
                self.pad_token,
                self.mask_token,
            ]
            if s
        ]

    @property
    def passed_bos_token(self) -> bool:
        """Return True if this token is in use."""
        return self.bos_token is not None

    @property
    def passed_eos_token(self) -> bool:
        """Return True if this token is in use."""
        return self.eos_token is not None

    @property
    def passed_unk_token(self) -> bool:
        """Return True if this token is in use."""
        return self.unk_token is not None

    @property
    def passed_pad_token(self) -> bool:
        """Return True if this token is in use."""
        return self.pad_token is not None

    @property
    def passed_mask_token(self) -> bool:
        """Return True if this token is in use."""
        return self.mask_token is not None

    @property
    def max_tokenized_length(self) -> int:
        """Return how long the tokenized list can be given the tokens used."""
        return self.max_length - (int(self.passed_bos_token) + int(self.passed_eos_token))

    @property
    def bos_token_int(self) -> int:
        """Return the integer encoding of the passed token."""
        if not self.passed_bos_token:
            raise ValueError(f"bos_token is false-y ({self.bos_token}), cannot get int.")
        return self.vocabulary.token_to_int[cast(str, self.bos_token)]

    @property
    def eos_token_int(self) -> int:
        """Return the integer encoding of the passed token."""
        if not self.passed_eos_token:
            raise ValueError(f"eos_token is false-y ({self.eos_token}), cannot get int.")
        return self.vocabulary.token_to_int[cast(str, self.eos_token)]

    @property
    def unk_token_int(self) -> int:
        """Return the integer encoding of the passed token."""
        if not self.passed_unk_token:
            raise ValueError(f"unk_token is false-y ({self.unk_token}), cannot get int.")
        return self.vocabulary.token_to_int[cast(str, self.unk_token)]

    @property
    def pad_token_int(self) -> int:
        """Return the integer encoding of the passed token."""
        if not self.passed_pad_token:
            raise ValueError(f"pad_token is false-y ({self.pad_token}), cannot get int.")
        return self.vocabulary.token_to_int[cast(str, self.pad_token)]

    @property
    def mask_token_int(self) -> int:
        """Return the integer encoding of the passed token."""
        if not self.passed_mask_token:
            raise ValueError(f"mask_token is false-y ({self.mask_token}), cannot get int.")
        return self.vocabulary.token_to_int[cast(str, self.mask_token)]


class SequenceTokenizer:
    """A sequence tokenizer."""

    def __init__(self, tokenizer_spec: SequenceTokenizerSpec):
        """Init the SequenceTokenizer class.

        Args:
            tokenizer_spec: The spec for the tokenizer.

        """
        self.tokenizer_spec = tokenizer_spec

    def _kmer_n(self, seq: str) -> List[str]:
        seq_len = len(seq)
        iterations = seq_len - self.tokenizer_spec.kmer_size + 1
        kmers = []

        for i in range(0, iterations, self.tokenizer_spec.kmer_step_size):
            kmer = seq[i : i + self.tokenizer_spec.kmer_size]
            kmers.append(kmer)

        return kmers

    def __call__(self, seq: str) -> List[str]:
        """Tokenize the seqeunce, see .tokenize."""
        return self.tokenize(seq)

    def encode(self, seq: str) -> List[int]:
        """Encode the underlying sequence."""
        return [self.tokenizer_spec.vocabulary.token_to_int[s] for s in self.tokenize(seq)]

    def tokenize(self, seq: str) -> List[str]:
        """Tokenize the sequence.

        Args:
            seq: The sequence to encode.

        Returns:
            The list of strs that are the tokens.

        """
        seq_len = len(seq)

        if seq_len < self.tokenizer_spec.kmer_size:
            raise ValueError(
                f"seq length {seq_len} cannot be less than the kmer "
                "size {self.tokenizer_spec.kmer_size}"
            )

        if self.tokenizer_spec.kmer_size == 1:
            kmer_list = list(seq)
        else:
            kmer_list = self._kmer_n(seq)

        sequence_kmers = kmer_list[: self.tokenizer_spec.max_tokenized_length]

        if self.tokenizer_spec.passed_bos_token:
            sequence_kmers.insert(0, cast(str, self.tokenizer_spec.bos_token))

        if self.tokenizer_spec.passed_eos_token:
            sequence_kmers.append(cast(str, self.tokenizer_spec.eos_token))

        return sequence_kmers



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved


import unittest

from click.testing import CliRunner

from gcgc import cli


class TestGcgc(unittest.TestCase):
    """Tests for `gcgc` package."""

    def test_command_line_interface(self):
        """Test the CLI."""
        runner = CliRunner()
        result = runner.invoke(cli.main)
        assert result.exit_code == 0
        help_result = runner.invoke(cli.main, ["--help"])
        assert help_result.exit_code == 0
        assert "--help  Show this message and exit." in help_result.output



# (c) Copyright 2019 Trent Hauck
# All Rights Reserved
"""Stores the alphabets, these are the same as BioPython."""

# pylint: disable=invalid-name

protein_letters = "ACDEFGHIKLMNPQRSTVWY"
extended_protein_letters = "ACDEFGHIKLMNPQRSTVWYBXZJUO"
ambiguous_dna_letters = "GATCRYWSMKHBVDN"

unambiguous_dna_letters = "GATC"
extended_dna_letters = "GATCBDSW"

ambiguous_rna_letters = "GAUCRYWSMKHBVDN"
unambiguous_rna_letters = "GAUC"



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved
"""Top-level GCGC module."""

__version__ = "0.11.0"

from gcgc.tokenizer import SequenceTokenizer
from gcgc.tokenizer import SequenceTokenizerSpec



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved
"""Test the alphabet."""

import pytest

from gcgc.tokenizer import SequenceTokenizer, SequenceTokenizerSpec


@pytest.mark.parametrize(
    "seq,spec,expected_tokens,expected_encoding",
    [
        (
            "ATCG",
            SequenceTokenizerSpec(alphabet="ATCG", max_length=10, kmer_size=1, kmer_step_size=1),
            list("ATCG"),
            [0, 1, 2, 3],
        ),
        (
            "ATCG",
            SequenceTokenizerSpec(alphabet="ATCG", max_length=2, kmer_size=1, kmer_step_size=1),
            list("AT"),
            [0, 1],
        ),
        (
            "ATCG",
            SequenceTokenizerSpec(alphabet="ATCG", max_length=10, kmer_size=2, kmer_step_size=1),
            ["AT", "TC", "CG"],
            [1, 6, 11],
        ),
        (
            "AATT",
            SequenceTokenizerSpec(alphabet="ATCG", max_length=10, kmer_size=2, kmer_step_size=2),
            ["AA", "TT"],
            [0, 5],
        ),
        (
            "ATCGATCGATCGATCG",
            SequenceTokenizerSpec(alphabet="ATCG", max_length=2, kmer_size=2, kmer_step_size=1),
            ["AT", "TC"],
            [1, 6],
        ),
        (
            "ATCGATCG",
            SequenceTokenizerSpec(
                alphabet="ATCG",
                bos_token=">",
                eos_token="<",
                max_length=10,
                kmer_size=2,
                kmer_step_size=1,
            ),
            [">", "AT", "TC", "CG", "GA", "AT", "TC", "CG", "<"],
            [0, 3, 8, 13, 14, 3, 8, 13, 1],
        ),
    ],
)
def test_kmer_tokenization(seq, spec, expected_tokens, expected_encoding):
    """Test the kemrs are encoded as expected."""
    tokenizer = SequenceTokenizer(spec)
    actual_tokens = tokenizer(seq)
    assert actual_tokens == expected_tokens

    actual_encoded = tokenizer.encode(seq)
    assert actual_encoded == expected_encoding



# (c) Copyright 2019 Trent Hauck
# All Rights Reserved
"""Sets up the GCGC package."""

from pathlib import Path

from setuptools import setup, find_packages

DESCRIPTION = "GCGC is a preprocessing library for biological sequence model development."
THIS_DIR = Path(__file__).resolve().parent


def _get_long_description():
    history = (THIS_DIR / "CHANGELOG.md").read_text()
    return (THIS_DIR / "README.md").read_text() + "\n\n" + history


setup(
    name="gcgc",
    version="0.11.0",
    description=DESCRIPTION,
    long_description=_get_long_description(),
    long_description_content_type="text/markdown",
    classifiers=["License :: OSI Approved :: MIT License"],
    author="Trent Hauck",
    author_email="trent@trenthauck.com",
    url="https://github.com/tshauck/gcgc",
    license="MIT",
    packages=find_packages(),
    install_requires=["pydantic~=1.1"],
    python_requires=">=3.7",
)



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved
"""Contains tokenizer level packages."""

from gcgc.tokenizer.base import SequenceTokenizer
from gcgc.tokenizer.base import SequenceTokenizerSpec

__all__ = ["SequenceTokenizer", "SequenceTokenizerSpec"]



# (c) Copyright 2018 Trent Hauck
# All Rights Reserved
"""CLI Functions for GCGC."""

import logging

import click

from gcgc import __version__

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


@click.group()
def main():
    """Console script for gcgc."""


@main.command()
def version():
    """Print the version and exit."""
    print(__version__)
